/*
 * 
 * SOCLIB_LGPL_HEADER_BEGIN
 * 
 * This file is part of SoCLib, GNU LGPLv2.1.
 * 
 * SoCLib is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation; version 2.1 of the License.
 * 
 * SoCLib is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with SoCLib; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA
 * 
 * SOCLIB_LGPL_HEADER_END
 * 
 * Copyright (c) CITI/INSA, 2009
 * 
 * Authors:
 * 	Ludovic L'Hours <ludovic.lhours@insa-lyon.fr>
 * 	Tanguy Risset <tanguy.risset@insa-lyon.fr>
 * 
 */

#include <vci_fir16_dma.h>
#include <read_to_fifo_request.h>
#include <write_from_fifo_request.h>

namespace soclib { namespace caba {

#define tmpl(x) \
template < typename vci_param > \
x VciFir16Dma<vci_param>

tmpl(bool)::on_write(int seg, typename vci_param::addr_t addr,
		typename vci_param::data_t data, int be) {

	unsigned int cell = addr / sc_dt::sc_uint<vci_param::N>(vci_param::B);

#ifdef SOCLIB_MODULE_DEBUG
	std::cout << name() << ": target write at " << std::hex << cell
			<< " data=" << data << " be=" << be << std::endl;
#endif

	if (cell >= FIR16_COEF0) {
		m_coeffs[cell - FIR16_COEF0] = fix_int<10>(data, true);
		return true;
	}

	switch (cell) {

		case FIR16_FROM:
			m_from = data;
			return true;

		case FIR16_TO:
			m_to = data;
			return true;

		case FIR16_LENGTH:
			m_length = ((unsigned int)data) * 2; // XXX words are short
			p_irq = false;
			if (m_length > 0) {
				m_dma_running = true;
				m_current_is_read = true;
				m_current_is_low = true;
				m_in_transfert = false;
#ifdef SOCLIB_MODULE_DEBUG
				std::cout << name() << ": start DMA" << std::endl;
#endif
			}
			return true;
	}
	return false;
}

tmpl(bool)::on_read(int seg, typename vci_param::addr_t addr,
		typename vci_param::data_t &data) {

	unsigned int cell = addr / sc_dt::sc_uint<vci_param::N>(vci_param::B);

#ifdef SOCLIB_MODULE_DEBUG
	std::cout << name() << ": target read at address "
			<< std::hex << addr << std::endl;
#endif

	switch (cell) {

#if 0
		case FIR16_OUTPUT:
			data = outData;
			return true;
#endif

	}
	return false;
}

tmpl(signed short)::doOneCycle(signed short data) {
	fix_int<10> inData = data;
	fix_int<10> sum = 0;
	for (int j = 15; j >= 0; --j) {
		sum += m_fir[j] * m_coeffs[j];
		m_fir[j] = j == 0 ? inData : m_fir[j-1];
	}
#ifdef SOCLIB_MODULE_DEBUG
	std::cout << name() << ": in = " << std::dec << data
			<< "         out = " << sum << " | "
			<< ((signed short)sum) << std::endl;
#endif
	return (signed short)sum;
}

tmpl(void)::next_read_request() {
	unsigned int current_count = m_length < m_burst ? m_length : m_burst;
	unsigned int available =
		(m_input_fifo.size() - m_input_fifo.filled_status()) * vci_param::B;
	if (current_count > available)
		current_count = available;
	if (current_count == 0)
		return;
#ifdef SOCLIB_MODULE_DEBUG
	std::cout << name() <<": next_read_request " << current_count << std::endl;
#endif
	ReadToFifoRequest<vci_param> *req =
			new ReadToFifoRequest<vci_param>(m_from, current_count);
	req->setOnWrite(this, ON_WRITE_FIFO(on_write_fifo));
	req->setDone(this, ON_T(read_request_done));
	m_in_transfert = true;
	m_vci_init_fsm.doReq(req);
}

tmpl(bool)::on_write_fifo(data_t data) {
	if (m_input_fifo.wok() == false)
		return false;
#ifdef SOCLIB_MODULE_DEBUG
	std::cout << name() << ": read_data " << std::hex << data << std::endl;
#endif
	m_input_write_bool = true;
	m_input_write_data = data;
	return true;
}

tmpl(void)::read_request_done(VciInitiatorReq<vci_param> *req) {
	m_in_transfert = false;
	if ( !req->failed() ) {
		unsigned int count = m_length < m_burst ? m_length : m_burst;
#ifdef SOCLIB_MODULE_DEBUG
		std::cout << name() <<": end read " << m_from << std::endl;
#endif
		m_length -= count;
		m_from += count;
	} else {
		std::cout << name() << ": DMA read request failed" << std::endl;
	}
	delete req;
}

tmpl(void)::next_write_request() {
	unsigned int current_count = m_output_fifo.filled_status() * vci_param::B;
	if (current_count > m_burst)
		current_count = m_burst;
	if (current_count == 0) {
		if (m_length == 0) {
#ifdef SOCLIB_MODULE_DEBUG
			std::cout << name() <<": DMA stop " << std::endl;
#endif
			m_dma_running = false;
			p_irq = true;
		}
		return;
	}
#ifdef SOCLIB_MODULE_DEBUG
	std::cout << name() <<": next_write_request " << current_count << std::endl;
#endif
	WriteFromFifoRequest<vci_param> *req =
			new WriteFromFifoRequest<vci_param>(m_to, current_count);
	req->setOnRead(this, ON_READ_FIFO(on_read_fifo));
	req->setDone(this, ON_T(write_request_done));
	m_in_transfert = true;
	m_vci_init_fsm.doReq(req);
}

tmpl(bool)::on_read_fifo(data_t& data) {
	if (m_output_fifo.rok() == false)
		return false;
	data = m_output_fifo.read();
	m_output_read_bool = true;
	return true;
}

tmpl(void)::write_request_done(VciInitiatorReq<vci_param> *req) {
	m_in_transfert = false;
	if ( !req->failed() ) {
		WriteFromFifoRequest<vci_param> *r =
				(WriteFromFifoRequest<vci_param>*) req;
		unsigned int count = r->length();
		m_to += count;
	} else {
		std::cout << name() << ": DMA write request failed" << std::endl;
	}
	delete req;
}

tmpl(void)::next_request() {
	if (! m_in_transfert) {
		if (m_current_is_read) {
			next_read_request();
			m_current_is_read = false;
		} else {
			next_write_request();
			if (m_length > 0)
				m_current_is_read = true;
		}
	}
}

tmpl(void)::transition () {

	if (p_resetn == false) {
		m_vci_target_fsm.reset();
		m_vci_init_fsm.reset();
		m_length = 0;
		m_dma_running = false;
		m_current_is_read = true;
		m_current_is_low = true;
		m_in_transfert = false;
		p_irq = false;
		for (int i = 0; i < 16; ++i) {
			m_fir[i] = 0;
		}
		return;
	}

	m_input_read_bool = false;
	m_input_write_bool = false;
	m_output_read_bool = false;
	m_output_write_bool = false;

	m_vci_target_fsm.transition();
	m_vci_init_fsm.transition();

	if (m_dma_running) {
		next_request();
		if (m_current_is_low) {
			if (m_input_fifo.rok()) {
				m_input_data = m_input_fifo.read();
				m_input_read_bool = true;
				m_current_is_low = false;
				m_output_data = doOneCycle(m_input_data & 0xFFFF) & 0xFFFF;
			}
		} else {
			if (m_output_fifo.wok()) {
				m_current_is_low = true;
				m_output_write_bool = true;
				m_output_data |= doOneCycle((m_input_data>>16) & 0xFFFF)<<16;
				m_output_write_data = m_output_data;
			}
		}
	}

	if (m_input_read_bool == true && m_input_write_bool == true) {
#ifdef SOCLIB_MODULE_DEBUG
		std::cout << name() <<": input put_and_get " << std::endl;
#endif
		m_input_fifo.put_and_get(m_input_write_data);
 	} else if (m_input_read_bool == false && m_input_write_bool == true) {
#ifdef SOCLIB_MODULE_DEBUG
		std::cout << name() <<": input simple_put " << std::endl;
#endif
		m_input_fifo.simple_put (m_input_write_data);
	} else if (m_input_read_bool == true && m_input_write_bool == false) {
#ifdef SOCLIB_MODULE_DEBUG
		std::cout << name() <<": input simple_get " << std::endl;
#endif
		m_input_fifo.simple_get();
	}

	if (m_output_read_bool == true && m_output_write_bool == true) {
#ifdef SOCLIB_MODULE_DEBUG
		std::cout << name() <<": output put_and_get " << std::endl;
#endif
		m_output_fifo.put_and_get(m_output_write_data);
 	} else if (m_output_read_bool == false && m_output_write_bool == true) {
#ifdef SOCLIB_MODULE_DEBUG
		std::cout << name() <<": output simple_put " << std::endl;
#endif
		m_output_fifo.simple_put (m_output_write_data);
	} else if (m_output_read_bool == true && m_output_write_bool == false) {
#ifdef SOCLIB_MODULE_DEBUG
		std::cout << name() <<": output simple_get " << std::endl;
#endif
		m_output_fifo.simple_get();
	}

}

tmpl(void)::genMoore () {

	m_vci_target_fsm.genMoore();
	m_vci_init_fsm.genMoore();

}

tmpl(/**/)::VciFir16Dma(sc_module_name insname, const MappingTable& mt,
		const IntTab& srcid, const IntTab& tgtid, int burst_size)
	: BaseModule(insname),
	m_input_fifo("input_fifo"), m_output_fifo("output_fifo"),
	m_vci_target_fsm(p_vci_target, mt.getSegmentList(tgtid)),
	m_vci_init_fsm(p_vci_initiator, mt.indexForId(srcid)) {

	m_vci_target_fsm.on_read_write(on_read, on_write);

	SC_METHOD (transition);
	dont_initialize();
	sensitive << p_clk.pos();

	SC_METHOD (genMoore);
	dont_initialize();
	sensitive << p_clk.neg();

	m_burst = burst_size;
	m_burst = (m_burst + vci_param::B) & ~vci_param::B;
	if (m_burst > FIFO_SIZE*vci_param::B) {
		m_burst = FIFO_SIZE*vci_param::B;
	}

#ifdef SOCLIB_DEBUG_MODULE
	std::cout << name() << ": burst=" << m_burst << std::endl;
#endif

}

tmpl(/**/)::~VciFir16Dma() {
}

}} // end of soclib::caba


/// Local Variables:
/// mode: hs-minor
/// c-basic-offset: 4
/// End:
